//
// Bmp2Cvi by ssjx (http://ssjx.co.uk)
//
// Compiles with: 
//		dmd -i bmp2cvi.d
//
// Version 2.4 D (25/02/23)
// 25/02/2023 - Source clean up, now Avi2Cvc/Bmp2Cvi share a lot of functions
// 24/02/2023 - Reduced 8Bit allocation
// 24/02/2023 - Had forgotten to write popular byte!
// 24/02/2023 - Reduced 2bit allocation
//
// Version 2.3 D (22/03/20)
// 
//


import std.stdio;
import std.File;
import std.format;

import std.bitmanip;	// swapEndian

import std.datetime.stopwatch;

import cv1,utils;


// Bitmap Header - 14 bytes (NOT 16)
struct Mybmpheader
{
	align(1):
	short type;
	int siz;
	short res1;
	short res2;
	int offset;
};

// Bitmap Info Header 40 bytes
struct Mybmpinfo
{
	align(1):
	int siz;
	int width;
	int height;
	short planes;
	short depth;
	int comp;
	int isize;
	int xppm;
	int yppm;
	int clrused;
	int clrimp;
};


// 20 
struct Mycviheader
{
	align(1):
	char[4] id;
	short x;
	short y;
	char[4] format;
	short type;
	ubyte popular;
	ubyte spare;
	short tablesize;
	short size;
};

//

ubyte popular(ubyte[] pix){
	int[256] count;

	foreach(px;pix){
		count[px]++;
	}

	// find largest
	int big=count[0];  // number of occurences
    int bigc=0;

	for(int c=1;c<count.length;c++){
	    if (count[c]>big){
			big=count[c];  // number of occurences
			bigc=c;        // actual value
	    }
	}
	
	return cast(ubyte)bigc;
}

//


void main(string[] args)
{	

	string fn;
	int dither=0;
	
	writeln("Bmp2Cvc v2.4 D (25/02/23) by ssjx (http://ssjx.co.uk)");
	writeln("-----------------------------------------------------");
	
	if (args.length>1)
	{
		foreach(par;args)
		{
			if (par[0]=='-')
			{
				if (par=="-ordered"){dither=0;}
				if (par=="-diffuse"){dither=1;}
				if (par=="-nodither"){dither=2;}
			}
			else
			{
				if (par!=args[0])
				{
					// Not the exe or an option so probably a bmp file
					fn=par;
				}
			}
		}
	}
	else
	{
		writeln("Usage: ",args[0]," [option] myfile.bmp");
		writeln("Options: -nodither , -ordered (default), -diffuse");
		return;
	}

    if (exists(fn))
    {
		try
		{
			StopWatch sw;
			sw.start();
		
			Mybmpheader[1] bmheader;
			Mybmpinfo[1] bminfo;
			Mycviheader[1] cvi_hdr;
		
			ubyte[] data; 	/// The actual bitmap image 
			
			File file = File(fn, "rb");
	        file.rawRead(bmheader);
	        file.rawRead(bminfo);

	        writeln(fn," : ",bminfo[0].width," x ",bminfo[0].height," x ",bminfo[0].depth);
		
			bool abort=false;
			if ((bminfo[0].width*bminfo[0].height)>(320*240))
			{
				abort=true;
				writeln("Image size too big, images should be a maximum of 320x200.");
				writeln("(The width should also be divisible by 4)");
			}

			if ((bminfo[0].width*bminfo[0].height)<0)
			{
				abort=true;
				writeln("Image size too big, images should be a maximum of 320x200.");
				writeln("(The width should also be divisible by 4)");
			}

			if ((bminfo[0].comp)!=0)
			{
				abort=true;
				writeln("Images should be uncompressed (raw).");
			}

			if ((bminfo[0].depth)!=24)
			{
				abort=true;
				writeln("Images should be 24bit colour.");
			}


			// Assuming we have not aborted...

			if (abort==false)
			{
				// Now allocate space...
				ubyte[] mem; // File buffer 24bit 320x200x3
				mem.length=bminfo[0].width*bminfo[0].height*3;
				
				ubyte[] img; // Image buffer (8bit gray)
				img.length=bminfo[0].width*bminfo[0].height;
				
				int cy_frame_size = bminfo[0].width * bminfo[0].height/4;
								
				file.rawRead(mem);
				file.close(); 

				writeln("Converting 24bit -> 8Bit");
				grey(mem,img,bminfo[0].width,bminfo[0].height);
				
				switch(dither)
				{
					case 0:
						writeln("Ordered dithering");
						ordered(img,bminfo[0].width,bminfo[0].height);
					break;
					
					case 1:
						writeln("Diffusing");
						diffuse(img,bminfo[0].width,bminfo[0].height);
					break;
					
					case 2:
						writeln("No Dither");
						twobit(img,bminfo[0].width,bminfo[0].height);
					break;
					
					default:
						writeln("Ordered dithering");
						ordered(img,bminfo[0].width,bminfo[0].height);
					break;
				}
				
				writeln("Converting to 4 pixels per byte");
				
				ubyte[] cy;
				cy.length=cy_frame_size;
				
				frameconv(img,cy,bminfo[0].width,bminfo[0].height);

				writeln("Creating fake previous frame");
				
				immutable ubyte pbyte=popular(cy);
				
				ubyte[] prev_frame;
				prev_frame.length=cy.length;
				prev_frame[]=pbyte;
				
				set_as_previous(prev_frame);
				
				writeln("Compression: ");

				// See which compression method works the best!
				short maxsize=cast(short)((bminfo[0].width*bminfo[0].height)/4);
				
				int esize=encode_rle_size(cy);
				int usecomp=1;	
				int tmpsize=0;
			
				tmpsize=encode_delta1_size(cy);
				if (tmpsize<esize){
					esize=tmpsize;
					usecomp=2;
				}
				
				tmpsize=encode_delta2_size(cy);
				if (tmpsize<esize){
					esize=tmpsize;
					usecomp=3;
				}
			
				// In case compression just does not work (or not worth it)...
				if (esize>=maxsize){
					esize=maxsize;
					usecomp=0;
				}
				
				// Force 
				// esize=encode_delta2_size(cy,maxsize);
				// usecomp=3;
				
				// esize=encode_rle_size(cy,maxsize);
				// usecomp=1;	
				//
				
				
				ubyte[] cvc_data;
				cvc_data.length=esize;
				
				final switch(usecomp)
				{
				case 0:
					// Uncompressed frame
					writeln("+ RAW Size:", esize);
					encode_none(cy,cvc_data);
				break;
				
				case 1:
					// RLE compressed frame...
					writeln("+ RLE Size:", esize);
					encode_rle(cy,cvc_data);
				break;
				
				case 2:
					// Delta 1 compressed frame...
					writeln("+ Delta 1 Size:", esize);
					encode_delta1(cy,cvc_data);
				break;
				
				case 3:
					// Delta 2 compressed frame...
					writeln("+ Delta 2 Size:", esize);
					encode_delta2(cy,cvc_data);
				break;
				}
				
				cvi_hdr[0].id[0]='c';
				cvi_hdr[0].id[1]='v';
				cvi_hdr[0].id[2]='i';
				cvi_hdr[0].id[3]='m';
				
				cvi_hdr[0].x=cast(short)(bminfo[0].width);
				cvi_hdr[0].x=cvi_hdr[0].x.swapEndian;
				
				cvi_hdr[0].y=cast(short)(bminfo[0].height);
				cvi_hdr[0].y=cvi_hdr[0].y.swapEndian;
				
				cvi_hdr[0].format[0]='c';
				cvi_hdr[0].format[1]='v';
				cvi_hdr[0].format[2]='d';
				cvi_hdr[0].format[3]='1';
				
				cvi_hdr[0].type=cast(short)(usecomp);
				cvi_hdr[0].type=cvi_hdr[0].type.swapEndian;
				
				cvi_hdr[0].popular=pbyte;
				
				cvi_hdr[0].size=cast(short)(esize);
				cvi_hdr[0].size=cvi_hdr[0].size.swapEndian;

				string outfn=fn~".cvi";	// <-- Note squiggle used to join strings!

				File outfile = File(outfn, "wb");   	
				writeln("Saving as ",outfn,"...");
				
				outfile.rawWrite(cvi_hdr);
				outfile.rawWrite(cvc_data);
				outfile.close();
			}
			else
			{
				printf("Aborted conversion\n");
			}

			writeln("Finished.\n");
			long msecs = sw.peek.total!"msecs";
			writeln("Elapsed time: ",msecs,"ms");
			
	    }
        catch (FileException e)
		{
		    writeln("Exception caught opening input file!");
		}  


	
    }
    else
    {
        writeln("Could not find the file [",fn,"]...");
    }


return;
}
